/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id: Space.java 1293736 2012-02-26 02:29:01Z gadams $ */

package org.apache.fop.area.inline;

/**
 * Inline space area.
 * This is used for adding a inline space to the output.
 */
public class Space extends InlineArea {

    private static final long serialVersionUID = -8748265505356839796L;

    /**
     * Default constructor.
     */
    public Space() {
    }

}
