/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id: Anchor.java 1036179 2010-11-17 19:45:27Z spepping $ */

package org.apache.fop.area.inline;

/**
 * Anchor area for footnote or float.
 * Not sure if this is needed.
 */
public class Anchor extends InlineArea {

    private static final long serialVersionUID = 5227798744787823499L;

    // has a keep with adjacent area
    // has reference to associated footnote or float out-of-line area

}

