/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* $Id$ */

package org.apache.fop.afp.modca;

/**
 * Represents a rotation value
 *
 */
public enum Rotation {
    ROTATION_0(0),
    ROTATION_90(0x2D),
    ROTATION_180(0x5A),
    ROTATION_270(0x87);

    private final byte firstByte;

    public void writeTo(byte[] out, int offset) {
        out[offset] = firstByte;
        out[offset + 1] = (byte)0;
    }

    private Rotation(int firstByte) {
        this.firstByte = (byte) firstByte;
    }

    public byte getByte() {
        return firstByte;
    }
}
